import os
import unittest
from src.training.model_trainer import Trainer
from src.training.data_prep import DataPrep
from dotenv import load_dotenv
load_dotenv()

class TestTrainerManager(unittest.TestCase):
    """Test cases for the Trainer class."""

    def setUp(self) -> None:
        # Common setup for both test cases
        self.model_id = "openai/whisper-tiny"
        process = DataPrep(
            huggingface_token=os.environ.get("HF_TOKEN"),
            dataset_name="mozilla-foundation/common_voice_16_1",
            language_abbr="af",
            model_id=self.model_id,
            processing_task="transcribe",
            use_peft=False,
        )
        tokenizer, feature_extractor, feature_processor, model = process.prepare_model()

        # Load datasets
        self.dataset_streaming = process.load_dataset(
            feature_extractor, tokenizer, feature_processor, streaming=True,
            train_num_samples=10, test_num_samples=10
        )
        self.dataset_batch = process.load_dataset(
            feature_extractor, tokenizer, feature_processor, streaming=False,
            train_num_samples=10, test_num_samples=10
        )

        # Check if train/test samples exist in both streaming and batch datasets
        self._validate_dataset(self.dataset_streaming, "streaming")
        self._validate_dataset(self.dataset_batch, "batch")

        # Set up trainers for both streaming and batch datasets
        self.trainer_streaming = Trainer(
            language=["af"],
            huggingface_token=os.environ.get("HF_TOKEN"),
            model_id=self.model_id,
            dataset=self.dataset_streaming,
            model=model,
            feature_processor=feature_processor,
            feature_extractor=feature_extractor,
            tokenizer=tokenizer,
            wandb_api_key="",
            use_peft=False,
            processing_task="transcribe"
        )
        self.trainer_batch = Trainer(
            language =["af"],
            huggingface_token="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",
            model_id=self.model_id,
            dataset=self.dataset_batch,
            model=model,
            feature_processor=feature_processor,
            feature_extractor=feature_extractor,
            tokenizer=tokenizer,
            wandb_api_key="e0fda284061622e0f7858d6c684281d48fa05ecf",
            use_peft=False,
            processing_task="transcribe"
        )

        return super().setUp()

    def _validate_dataset(self, dataset, dataset_type):
        """Helper function to validate that datasets are not empty."""
        has_train_sample = any(True for _ in dataset["train"])
        assert has_train_sample, f"Train dataset for {dataset_type} is empty!"

        has_test_sample = any(True for _ in dataset["test"])
        assert has_test_sample, f"Test dataset for {dataset_type} is empty!"

    def test_01_train_streaming(self):
        """Test case for training with the streaming dataset."""
        self.trainer_streaming.train(
            max_steps=15,
            learning_rate=1e-5,
            save_steps=10,
            eval_steps=10,
            logging_steps=10,
            output_dir=f"../{self.model_id}-finetuned",
            report_to=None,
            push_to_hub=False,
            use_cpu=True,
            optim="adamw_hf",
            per_device_train_batch_size=4
        )
        # Check if output files exist after training
        assert os.path.exists(f"../{self.model_id}-finetuned/preprocessor_config.json")
        assert os.path.exists(f"../{self.model_id}-finetuned/tokenizer_config.json")

    def test_02_train_batch(self):
        """Test case for training with the batch dataset."""
        self.trainer_batch.train(
            max_steps=10,
            learning_rate=1e-5,
            save_steps=10,
            eval_steps=10,
            logging_steps=10,
            output_dir=f"../{self.model_id}-finetuned",
            report_to=None,
            push_to_hub=False,
            use_cpu=True,
            optim="adamw_hf"
        )
        # Check if output files exist after training
        assert os.path.exists(f"../{self.model_id}-finetuned/preprocessor_config.json")
        assert os.path.exists(f"../{self.model_id}-finetuned/tokenizer_config.json")


if __name__ == '__main__':
    unittest.main()



import os
import unittest
# from datasets import Dataset
from src.training.data_prep import DataPrep
from src.training.load_data import Dataset
from src.training.whisper_model_prep import WhisperModelPrep
from datasets import IterableDataset
from dotenv import load_dotenv
load_dotenv()

class TestDatasetManager(unittest.TestCase):
    """Test cases for the DatasetManager class."""

    def setUp(self):
        """Set up the testing environment."""
        self.data_prep = DataPrep(
            huggingface_token= os.environ.get("HF_TOKEN"),
            dataset_name="mozilla-foundation/common_voice_16_1",
            language_abbr="af",
            model_id="openai/whisper-small",
            processing_task="transcribe",
            use_peft=False,
        )
        self.model_prep=WhisperModelPrep(
            language= "af",
            model_id="openai/whisper-small",
            processing_task="transcribe",
            use_peft=False),
        self.data_loader=Dataset(os.environ.get("HF_TOKEN"), "mozilla-foundation/common_voice_16_1", ["yi", "ti"])

    def test_load_dataset_streaming_true(self):
        """Test the load_dataset method."""
        tokenizer, feature_extractor, processor, model = self.data_prep.prepare_model()
        dataset = self.data_prep.load_dataset(feature_extractor, tokenizer, processor, train_num_samples = 10, test_num_samples=10)
        has_train_sample = any(True for _ in dataset["train"])
        assert has_train_sample, "Train dataset is empty!"

        has_test_sample = any(True for _ in dataset["test"])
        assert has_test_sample, "Test dataset is empty!"
        self.assertIsInstance(dataset, dict)
        self.assertIsNotNone(dataset["train"])
        self.assertIsNotNone(dataset["test"])
        self.assertIsInstance(dataset["train"], IterableDataset)
        self.assertIsInstance(dataset["test"], IterableDataset)


    def test_load_dataset_streaming_false(self):
            """Test the load_dataset method."""
            tokenizer, feature_extractor, processor, model = self.data_prep.prepare_model()
            dataset = self.data_prep.load_dataset(feature_extractor, tokenizer, processor, streaming=False, train_num_samples=10,
                                                  test_num_samples=10)
            has_train_sample = any(True for _ in dataset["train"])
            assert has_train_sample, "Train dataset is empty!"

            has_test_sample = any(True for _ in dataset["test"])
            assert has_test_sample, "Test dataset is empty!"
            self.assertIsInstance(dataset, dict)
            self.assertIsNotNone(dataset["train"])
            self.assertIsNotNone(dataset["test"])
            # self.assertIsInstance(dataset["train"], IterableDataset)
            # self.assertIsInstance(dataset["test"], IterableDataset)



if __name__ == '__main__':
    unittest.main()



import os
from transformers import Seq2SeqTrainingArguments, Seq2SeqTrainer, WhisperForConditionalGeneration
from .collator import DataCollatorSpeechSeq2SeqWithPadding, DataCollatorSpeechSeq2SeqWithPaddingPS
import evaluate
import torch
import time
import csv
from tqdm import tqdm
from huggingface_hub import upload_folder
from torch.utils.data import DataLoader
from .wandb_callback import WandbProgressResultsCallback
from transformers.trainer_pt_utils import IterableDatasetShard
from torch.utils.data import IterableDataset
from transformers import TrainerCallback
from .whisper_model_prep import WhisperModelPrep
from typing import Dict, Any
from .evaluation import log_pred, log_metric, compute_metrics
from accelerate.logging import get_logger
import warnings

warnings.filterwarnings("ignore")
logger = get_logger(__name__)


class ShuffleCallback(TrainerCallback):
    def on_epoch_begin(self, args, state, control, train_dataloader, **kwargs):
        if isinstance(train_dataloader.dataset, IterableDatasetShard):
            pass
        elif isinstance(train_dataloader.dataset, IterableDataset):
            train_dataloader.dataset.set_epoch(train_dataloader.dataset._epoch + 1)


class Trainer:
    """

    A Trainer class for fine-tuning and training speech-to-text models using the Hugging Face Transformers library.

    """

    def __init__(
        self,
        huggingface_token: str,
        model_id: str,
        dataset: dict,
        language: list,
        model: WhisperForConditionalGeneration,
        feature_processor,
        feature_extractor,
        tokenizer,
        wandb_api_key: str,
        use_peft: bool,
        processing_task:str
    ):
        """
        Initializes the Trainer with the necessary components and configurations for training.

        Parameters:
            huggingface_token (str): Hugging Face API token for authenticated push access.
            model_id (str): Identifier for the pre-trained model.
            dataset (dict): The dataset split into 'train' and 'test'.
            model (PreTrainedModel): The model instance to be trained.
            feature_processor (Any): The audio feature processor.
            feature_extractor (Any): The audio feature extractor.
            tokenizer (PreTrainedTokenizer): The tokenizer for text data.
            processing_task (str): task for the Whisper model to execute. Translate or Transcribe
        """
        os.environ["WANDB_API_KEY"] = wandb_api_key
        self.dataset = dataset
        self.model = model
        self.model_id = model_id
        self.tokenizer = tokenizer
        self.feature_processor = feature_processor
        self.feature_extractor = feature_extractor
        self.huggingface_token = huggingface_token
        self.use_peft = use_peft
        self.model_prep = WhisperModelPrep(
            model_id=self.model_id,
            processing_task=processing_task,
            use_peft=self.use_peft,
            language=language
        )

    def compute_metrics(self, pred) -> dict:
        """
        Computes the Word Error Rate (WER) metric for the model predictions.

        Parameters:
            pred (PredictionOutput): The output from the model's prediction.

        Returns:
            dict: A dictionary containing the computed WER metric.
        """
        pred_ids = pred.predictions
        label_ids = pred.label_ids
        label_ids[label_ids == -100] = self.tokenizer.pad_token_id
        pred_str = self.tokenizer.batch_decode(
            pred_ids, skip_special_tokens=True, normalize=True
        )
        label_str = self.tokenizer.batch_decode(
            label_ids, skip_special_tokens=True, normalize=True
        )
        metric = evaluate.load("wer")
        wer = 100 * metric.compute(predictions=pred_str, references=label_str)
        return {"wer": wer}


    def compute_spectrograms(self, example: Dict[str, Any]) -> Dict[str, Any]:
        """
        Computes spectrograms from audio waveform.

        Args:
            example (dict): A dictionary containing audio waveform data.

        Returns:
            dict: A dictionary containing computed spectrogram data.

        Raises:
            KeyError: If the input example does not contain the required keys.
        """
        waveform = example["audio"]["array"]
        feature_extractor = self.model_prep.initialize_feature_extractor()
        specs = feature_extractor(
            waveform, sampling_rate=16000, padding="do_not_pad"
        ).input_features[0]
        return {"spectrogram": specs}

    def train(self,
        output_dir: str = None,
        max_steps: int = 100,
        learning_rate: float = 1e-5,
        lr_scheduler_type="constant_with_warmup",
        per_device_train_batch_size: int = 8,
        per_device_eval_batch_size: int = 8,
        optim: str = "adamw_bnb_8bit",
        gradient_accumulation_steps: int =1,
        gradient_checkpointing: bool = True,
        fp16: bool = torch.cuda.is_available(),
        evaluation_strategy: str = "steps",
        predict_with_generate: bool = True,
        generation_max_length: int = 225,
        save_steps: int = 25,
        eval_steps: int = 25,
        logging_steps: int = 25,
        load_best_model_at_end: bool = True,
        metric_for_best_model: str = "wer",
        greater_is_better: bool = False,
        push_to_hub: bool = True,
        hub_strategy: str = "checkpoint",
        save_safetensors: bool = True,
        resume_from_checkpoint: str = "last-checkpoint",
        report_to: str = "wandb",
        remove_unused_columns: bool = False,
        ignore_data_skip: bool = True,
        **kwargs):
        """
        Trains the model using the specified configurations.

        Args:
            output_dir (str): The output directory where the model predictions and checkpoints will be written.
            max_steps (int, optional): The maximum number of training steps. Defaults to 100.
            learning_rate (float, optional): The learning rate for the training process. Defaults to 1e-5.
            per_device_train_batch_size (int, optional): The batch size per GPU during training. Defaults to 8.
            per_device_eval_batch_size (int, optional): The batch size per GPU during evaluation. Defaults to 8.
            optim (str, optional): The optimizer to use for training. Defaults to "adamw_bnb_8bit".
            gradient_accumulation_steps (int, optional): The number of steps to accumulate gradients before performing an optimization step. Defaults to 1.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing to save memory at the expense of slower backward pass. Defaults to True.
            fp16 (bool, optional): Whether to use 16-bit (mixed) precision training instead of 32-bit training.
            evaluation_strategy (str, optional): The evaluation strategy to adopt during training. "steps": Evaluate every `eval_steps`. Defaults to "steps".
            predict_with_generate (bool, optional): Whether to use generate to calculate generative metrics (ROUGE, BLEU). Defaults to True.
            generation_max_length (int, optional): The maximum length of the sequence to be generated. Defaults to 225.
            save_steps (int, optional): The number of training steps before saving the model. Defaults to 25.
            eval_steps (int, optional): The number of training steps before evaluating the model. Defaults to 25.
            logging_steps (int, optional): The number of training steps before logging the training info. Defaults to 25.
            load_best_model_at_end (bool, optional): Whether to load the best model found during training at the end of training. Defaults to True.
            metric_for_best_model (str, optional): The metric to use to compare two different models. Defaults to "wer".
            greater_is_better (bool, optional): Whether a larger metric value indicates a better model. Defaults to False.
            push_to_hub (bool, optional): Whether to push the model to the Hugging Face model hub at the end of training. Defaults to True.
            hub_strategy (str, optional): The hub strategy to use for model checkpointing. Defaults to "checkpoint".
            save_safetensors (bool, optional): Whether to save tensors in a safe format. Defaults to False.
            resume_from_checkpoint (str, optional): The path to a checkpoint from which to resume training. Defaults to "last-checkpoint".
            report_to (str, optional): The list of integrations to report the results and logs to. Defaults to "wandb".
            remove_unused_columns (bool, optional): Whether to remove columns not required by the model when using a dataset. Defaults to False.
            ignore_data_skip (bool, optional): Whether to skip data loading issues when the dataset is being created. Defaults to True.
            **kwargs: Additional keyword arguments to be passed to the `Seq2SeqTrainingArguments` constructor https://huggingface.co/docs/transformers/main_classes/trainer#transformers.Seq2SeqTrainingArguments.
        """
        data_collator = DataCollatorSpeechSeq2SeqWithPadding(
            processor=self.feature_processor
        )
        output_dir = f"../{self.model_id}-finetuned"
        training_args = Seq2SeqTrainingArguments(
            output_dir=output_dir,
            per_device_train_batch_size=per_device_train_batch_size,
            gradient_accumulation_steps=gradient_accumulation_steps,
            learning_rate=learning_rate,
            lr_scheduler_type=lr_scheduler_type,
            max_steps=max_steps,
            gradient_checkpointing=gradient_checkpointing,
            fp16=fp16,
            optim=optim,
            evaluation_strategy=evaluation_strategy,
            per_device_eval_batch_size=per_device_eval_batch_size,
            predict_with_generate=predict_with_generate,
            generation_max_length=generation_max_length,
            save_steps=save_steps,
            eval_steps=eval_steps,
            logging_steps=logging_steps,
            load_best_model_at_end=load_best_model_at_end,
            metric_for_best_model=metric_for_best_model,
            greater_is_better=greater_is_better,
            push_to_hub=push_to_hub,
            hub_token=self.huggingface_token,
            hub_strategy = hub_strategy,
            save_safetensors = save_safetensors,
            resume_from_checkpoint  = resume_from_checkpoint,
            report_to=report_to,
            remove_unused_columns=remove_unused_columns,
            ignore_data_skip=ignore_data_skip,
            **kwargs
        )

        eval_dataset = self.dataset["test"].map(self.compute_spectrograms)

        trainer = Seq2SeqTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dataset["train"],
            eval_dataset=eval_dataset,
            data_collator=data_collator,
            compute_metrics=self.compute_metrics,
            tokenizer=self.feature_processor.feature_extractor,
            callbacks=[ShuffleCallback()],
        )
        data_loader = trainer.get_train_dataloader()
        for batch in data_loader:
            if batch is None or len(batch) == 0:
                print("Empty batch found!")
                break
            # print("Batch contains data")
        tokenizer = self.model_prep.initialize_tokenizer()
        processor = self.model_prep.initialize_processor()
        tokenizer.save_pretrained(training_args.output_dir)
        processor.save_pretrained(training_args.output_dir)
        progress_callback = WandbProgressResultsCallback(
            trainer, eval_dataset, tokenizer
        )
        trainer.add_callback(progress_callback)
        trainer.train()





###############################





def filter_eot_tokens(preds, decoder_eot_token_id):
    for idx in range(len(preds)):
        # remove the EOT tokens to get the 'true' token length
        token_ids = [token for token in preds[idx] if token != decoder_eot_token_id]
        token_ids = token_ids + [decoder_eot_token_id]
        preds[idx] = token_ids
    return preds


def training_schedule(
    processor,
    model,
    max_label_length,
    language,
    task,
    accelerator,
    vectorized_datasets,
    file_ids_dataset,
    output_dir,
    torch_dtype,
    tokenizer,
    push_to_hub,
    repo_name,
    raw_datasets,
    decoder_eot_token_id,
    decoder_prev_token_id,
    timestamp_position,
    data_splits,
    dataset_config_name,
    normalizer,
    logging_steps=None,
    preprocessing_batch_size=None,
    per_device_eval_batch_size=None,
    num_workers=2,
    dataloader_num_workers=None,
    concatenate_audio=True,
    streaming=False,
    generation_num_beams=None,
    return_timestamps=False
):

    per_device_eval_batch_size = int(per_device_eval_batch_size)

    data_collator = DataCollatorSpeechSeq2SeqWithPaddingPS(
        processor=processor,
        decoder_start_token_id=model.config.decoder_start_token_id,  # <|startoftranscript|>
        input_padding="longest",
        target_padding="max_length",
        max_target_length=max_label_length,
    )

    # 14. Define generation arguments - we need to do this before we wrap the models in DDP
    # so that we can still access the configs
    num_beams = (
        generation_num_beams
        if generation_num_beams is not None
        else getattr(model.generation_config, "num_beams", 1)
    )

    gen_kwargs = {
        "max_length": max_label_length,
        "num_beams": num_beams,
        "return_timestamps": return_timestamps,
    }
    if hasattr(model.generation_config, "is_multilingual") and model.generation_config.is_multilingual:
        # forcing the language and task tokens helps multilingual models in their generations
        gen_kwargs.update(
            {
                "language": language,
                "task": task,
            }
        )
    # remove any preset forced decoder ids since these are deprecated
    model.generation_config.forced_decoder_ids = None
    model.config.forced_decoder_ids = None

    # 15. Prepare everything with accelerate
    model = accelerator.prepare(model)

    def eval_step_with_save(split="eval"):
        # ======================== Evaluating ==============================
        eval_preds = []
        eval_labels = []
        eval_ids = []
        pred_str = []
        eval_start = time.time()

        eval_loader = DataLoader(
            vectorized_datasets[split],
            batch_size=per_device_eval_batch_size,
            collate_fn=data_collator,
            num_workers=dataloader_num_workers,
            pin_memory=True,
        )
        file_loader = DataLoader(
            file_ids_dataset[split],
            batch_size=per_device_eval_batch_size * accelerator.num_processes,
            num_workers=dataloader_num_workers,
        )

        eval_loader = accelerator.prepare(eval_loader)
        batches = tqdm(eval_loader, desc=f"Evaluating {split}...", disable=not accelerator.is_local_main_process)

        # make the split name pretty for librispeech etc
        split = split.replace(".", "-").split("/")[-1]
        output_csv = os.path.join(output_dir, f"{split}-transcription.csv")

        for step, (batch, file_ids) in enumerate(zip(batches, file_loader)):
            # Generate predictions and pad to max generated length
            generate_fn = model.module.generate if accelerator.num_processes > 1 else model.generate
            generated_ids = generate_fn(batch["input_features"].to(dtype=torch_dtype), **gen_kwargs)
            generated_ids = accelerator.pad_across_processes(generated_ids, dim=1, pad_index=tokenizer.pad_token_id)
            # Gather all predictions and targets
            generated_ids, labels = accelerator.gather_for_metrics((generated_ids, batch["labels"]))
            eval_preds.extend(generated_ids.cpu().numpy())
            eval_labels.extend(labels.cpu().numpy())
            eval_ids.extend(file_ids)

            if step % logging_steps == 0 and step > 0:
                batches.write(f"Saving transcriptions for split {split} step {step}")
                accelerator.wait_for_everyone()
                pred_ids = eval_preds[-(len(eval_preds) - len(pred_str)):]
                pred_ids = filter_eot_tokens(pred_ids, decoder_eot_token_id)
                pred_str.extend(
                    tokenizer.batch_decode(
                        pred_ids, skip_special_tokens=False, decode_with_timestamps=return_timestamps
                    )
                )
                csv_data = [[eval_ids[i], pred_str[i]] for i in range(len(eval_preds))]

                with open(output_csv, "w", encoding="UTF8", newline="") as f:
                    writer = csv.writer(f)
                    # write multiple rows
                    writer.writerow(["file_id", "whisper_transcript"])
                    writer.writerows(csv_data)

                if push_to_hub and accelerator.is_main_process:
                    upload_folder(
                        folder_path=output_dir,
                        repo_id=repo_name,
                        repo_type="dataset",
                        commit_message=f"Saving transcriptions for split {split} step {step}.",
                    )

        accelerator.wait_for_everyone()
        eval_time = time.time() - eval_start

        # compute WER metric for eval sets
        wer_desc = ""
        if "validation" in split or "test" in split:
            eval_preds = filter_eot_tokens(eval_preds, decoder_eot_token_id)
            wer_metric, pred_str, label_str, norm_pred_str, norm_label_str, eval_ids = compute_metrics(
                eval_preds, eval_labels, eval_ids, tokenizer, normalizer
            )
            wer_desc = " ".join([f"Eval {key}: {value} |" for key, value in wer_metric.items()])
            # Save metrics + predictions
            log_metric(
                accelerator,
                metrics=wer_metric,
                train_time=eval_time,
                prefix=split,
            )
            log_pred(
                accelerator,
                pred_str,
                label_str,
                norm_pred_str,
                norm_label_str,
                prefix=split,
            )
        else:
            pred_ids = eval_preds[-(len(eval_preds) - len(pred_str)):]
            pred_ids = filter_eot_tokens(pred_ids, decoder_eot_token_id)
            pred_str.extend(
                tokenizer.batch_decode(pred_ids, skip_special_tokens=False, decode_with_timestamps=return_timestamps)
            )

        batches.write(f"Saving final transcriptions for split {split}.")
        csv_data = [[eval_ids[i], eval_preds[i]] for i in range(len(eval_preds))]
        with open(output_csv, "w", encoding="UTF8", newline="") as f:
            writer = csv.writer(f)
            # write multiple rows
            writer.writerow(["file_id", "whisper_transcript"])
            writer.writerows(csv_data)

        # Print metrics
        logger.info(wer_desc)

        if not streaming:
            raw_datasets[split] = raw_datasets[split].add_column("whisper_transcript", pred_str)
            raw_datasets[split] = raw_datasets[split].add_column("eval_preds", eval_preds)

            def add_concatenated_text(eval_preds, condition_on_prev):
                concatenated_prev = [None]
                for token_ids, condition in zip(eval_preds[:-1], condition_on_prev[1:]):
                    if condition is False:
                        concatenated_prev.append(None)
                    else:
                        prompt_ids = [token for token in token_ids if token != decoder_eot_token_id]
                        prompt_ids = [decoder_prev_token_id] + prompt_ids[timestamp_position:]
                        concatenated_prev.append(prompt_ids)
                return {"condition_on_prev": concatenated_prev}

            if concatenate_audio:
                with accelerator.main_process_first():
                    raw_datasets[split] = raw_datasets[split].map(
                        add_concatenated_text,
                        input_columns=["eval_preds", "condition_on_prev"],
                        remove_columns=["eval_preds"],
                        desc="Setting condition on prev...",
                        batched=True,
                        batch_size=preprocessing_batch_size,
                        num_proc=num_workers,
                    )


    logger.info("***** Running Labelling *****")
    logger.info("  Instantaneous batch size per device =" f" {per_device_eval_batch_size}")
    logger.info(
        f"  Total eval batch size (w. parallel & distributed) = {per_device_eval_batch_size * accelerator.num_processes}"
    )
    logger.info(f"  Predict labels with timestamps = {return_timestamps}")
    for split in data_splits:
        eval_step_with_save(split=split)
        accelerator.wait_for_everyone()
        if push_to_hub and accelerator.is_main_process:
            upload_folder(
                folder_path=output_dir,
                repo_id=repo_name,
                repo_type="dataset",
                commit_message=f"Saving final transcriptions for split {split.replace('.', '-').split('/')[-1]}",
            )
    if not streaming and accelerator.is_main_process:
        raw_datasets.save_to_disk(output_dir, num_proc=num_workers)
        if push_to_hub:
            raw_datasets.push_to_hub(repo_name, config_name=dataset_config_name)
    accelerator.end_training()


import evaluate
import warnings
from typing import List, Dict
import numpy as np
warnings.filterwarnings("ignore")

class MetricComputer:
    """
    A class to compute metrics for model predictions.

    This class encapsulates the functionality of loading a specific metric from the
    `evaluate` library and computing it using predictions and labels processed by a tokenizer.

    """

    def __init__(self, metric_name: str, tokenizer):
        """
        Initializes the MetricComputer with a specified metric and tokenizer.

        Args:
            metric_name (str): The name of the metric to load (e.g., "wer" for word error rate).
            tokenizer: The tokenizer to use for decoding prediction and label IDs.
        """
        self.metric = evaluate.load(metric_name)
        self.tokenizer = tokenizer

    def compute_metrics(self, pred) -> dict:
        """
        Computes the metric for a set of predictions and labels.

        The function processes the predictions and labels, replacing any padding token IDs with -100,
        and decodes them using the provided tokenizer before computing the metric.

        Args:
            pred: An object that contains `predictions` and `label_ids`, which are used to compute the metric.
                  The `predictions` and `label_ids` should be arrays of IDs.

        Returns:
            dict: A dictionary containing the computed metric value.
        """
        pred_ids = pred.predictions
        label_ids = pred.label_ids
        label_ids[label_ids == -100] = self.tokenizer.pad_token_id
        pred_str = self.tokenizer.batch_decode(pred_ids, skip_special_tokens=True)
        label_str = self.tokenizer.batch_decode(label_ids, skip_special_tokens=True)
        wer = 100 * self.metric.compute(predictions=pred_str, references=label_str)

        return {"wer": wer}

###############################

def log_metric(
    accelerator,
    metrics: Dict,
    train_time: float,
    prefix: str = "eval",
):
    """Helper function to log all evaluation metrics with the correct prefixes and styling."""
    log_metrics = {}
    for k, v in metrics.items():
        log_metrics[f"{prefix}/{k}"] = v
    log_metrics[f"{prefix}/time"] = train_time
    accelerator.log(log_metrics)


def log_pred(
    accelerator,
    pred_str: List[str],
    label_str: List[str],
    norm_pred_str: List[str],
    norm_label_str: List[str],
    prefix: str = "eval",
    num_lines: int = 200000,
):
    """Helper function to log target/predicted transcriptions to weights and biases (wandb)."""
    if accelerator.is_main_process:
        wandb_tracker = accelerator.get_tracker("wandb")
        # pretty name for split
        prefix = prefix.replace("/", "-")

        # convert str data to a wandb compatible format
        str_data = [[label_str[i], pred_str[i], norm_label_str[i], norm_pred_str[i]] for i in range(len(pred_str))]
        # log as a table with the appropriate headers
        wandb_tracker.log_table(
            table_name=f"{prefix}/all_predictions",
            columns=["Target", "Pred", "Norm Target", "Norm Pred"],
            data=str_data[:num_lines],
        )

        # log incorrect normalised predictions
        str_data = np.asarray(str_data)
        str_data_incorrect = str_data[str_data[:, -2] != str_data[:, -1]]
        # log as a table with the appropriate headers
        wandb_tracker.log_table(
            table_name=f"{prefix}/incorrect_predictions",
            columns=["Target", "Pred", "Norm Target", "Norm Pred"],
            data=str_data_incorrect[:num_lines],
        )


def compute_metrics(
        preds,
        labels,
        file_ids,
        tokenizer,
        normalizer,
        return_timestamps=False
):
    # replace padded labels by the padding token
    for idx in range(len(labels)):
        labels[idx][labels[idx] == -100] = tokenizer.pad_token_id

    pred_str = tokenizer.batch_decode(preds, skip_special_tokens=False, decode_with_timestamps=return_timestamps)
    # we do not want to group tokens when computing the metrics
    label_str = tokenizer.batch_decode(labels, skip_special_tokens=True)

    # normalize everything and re-compute the WER
    norm_pred_str = [normalizer(pred) for pred in pred_str]
    norm_label_str = [normalizer(label) for label in label_str]
    # for logging, we need the pred/labels to match the norm_pred/norm_labels, so discard any filtered samples here
    pred_str = [pred_str[i] for i in range(len(norm_pred_str)) if len(norm_label_str[i]) > 0]
    label_str = [label_str[i] for i in range(len(norm_label_str)) if len(norm_label_str[i]) > 0]
    file_ids = [file_ids[i] for i in range(len(file_ids)) if len(norm_label_str[i]) > 0]
    # filtering step to only evaluate the samples that correspond to non-zero normalized references:
    norm_pred_str = [norm_pred_str[i] for i in range(len(norm_pred_str)) if len(norm_label_str[i]) > 0]
    norm_label_str = [norm_label_str[i] for i in range(len(norm_label_str)) if len(norm_label_str[i]) > 0]
    metric = evaluate.load("wer")
    wer = 100 * metric.compute(predictions=norm_pred_str, references=norm_label_str)

    return {"wer": wer}, pred_str, label_str, norm_pred_str, norm_label_str, file_ids


from .asr import WhisperModel, FasterWhisperPipeline
from typing import Optional
import faster_whisper
import torch
from .vad import load_vad_model



def load_asr_model(whisper_arch,
               device,
               device_index=0,
               compute_type="float16",
               asr_options=None,
               language : Optional[str] = None,
               vad_model=None,
               vad_options=None,
               model : Optional[WhisperModel] = None,
               download_root=None,
               threads=4,
               is_v3_architecture=False):
    """
    Loads and returns a Whisper ASR model for inference.

    Args:
        whisper_arch (str): Name of the Whisper model architecture to load.
        device (str): Device to load the model on (e.g., "cuda" or "cpu").
        device_index (int): Index of the device, if applicable. Default is 0.
        compute_type (str): Type of compute precision to use (e.g., "float16").
        asr_options (dict, optional): Additional ASR-specific options.
        language (str, optional): Language code for the model. Defaults to None.
        vad_model (optional): Voice Activity Detection model instance, if used.
        vad_options (dict, optional): Options for VAD, if applicable.
        model (WhisperModel, optional): Preloaded WhisperModel instance, if available.
        download_root (str, optional): Directory path for downloading model files.
        threads (int): Number of CPU threads per worker. Default is 4.
        is_v3_architecture (bool): Indicates whether the model uses the v3 architecture.

    Returns:
        object: The Whisper model pipeline for ASR inference.
    """

    if whisper_arch.endswith(".en"):
        language = "en"

    model = model or WhisperModel(whisper_arch,
                         device=device,
                         device_index=device_index,
                         compute_type=compute_type,
                         download_root=download_root,
                         cpu_threads=threads)

    if is_v3_architecture:
        model.feature_extractor.mel_filters = model.feature_extractor.get_mel_filters(
            model.feature_extractor.sampling_rate,
            model.feature_extractor.n_fft,
            n_mels=128
        )
    else:
        model.feature_extractor.mel_filters = model.feature_extractor.get_mel_filters(
            model.feature_extractor.sampling_rate,
            model.feature_extractor.n_fft,
            n_mels=80
        )

    if language is not None:
        tokenizer = faster_whisper.tokenizer.Tokenizer(model.hf_tokenizer, model.model.is_multilingual, language=language)
    else:
        print("No language specified, language will be first be detected for each audio file (increases inference time).")
        tokenizer = None

    default_asr_options =  { # explore temperature_increment_on_fallback parameter
        "beam_size": 5,
        "best_of": 5,
        "patience": 1,
        "length_penalty": 1,
        "repetition_penalty": 1,
        "no_repeat_ngram_size": 0,
        "temperatures": [0.0, 0.2, 0.4, 0.6, 0.8, 1.0],
        "compression_ratio_threshold": 2.4,
        "log_prob_threshold": -1.0,
        "no_speech_threshold": 0.6,
        "condition_on_previous_text": False, # explore True
        "prompt_reset_on_temperature": 0.5,
        "initial_prompt": None,
        "prefix": None,
        "suppress_blank": True,
        "suppress_tokens": [-1],
        "without_timestamps": True,
        "max_initial_timestamp": 0.0,
        "word_timestamps": False, # Explore True
        "prepend_punctuations": "\"'“¿([{-",
        "append_punctuations": "\"'.。,，!！?？:：”)]}、",
        "suppress_numerals": False,
        "max_new_tokens": None,
        "clip_timestamps": None,
        "hallucination_silence_threshold": None,
        "hotwords": None
    }

    if asr_options is not None:
        default_asr_options.update(asr_options)

    suppress_numerals = default_asr_options["suppress_numerals"]
    del default_asr_options["suppress_numerals"]

    default_asr_options = faster_whisper.transcribe.TranscriptionOptions(**default_asr_options)

    default_vad_options = {
        "vad_onset": 0.500,
        "vad_offset": 0.363
    }

    if vad_options is not None:
        default_vad_options.update(vad_options)

    if vad_model is not None:
        vad_model = vad_model
    else:
        vad_model = load_vad_model(torch.device(device), use_auth_token=None, **default_vad_options)

    return FasterWhisperPipeline(
        model=model,
        vad=vad_model,
        options=default_asr_options,
        tokenizer=tokenizer,
        language=language,
        suppress_numerals=suppress_numerals,
        vad_params=default_vad_options,
        is_v3_architecture=is_v3_architecture
    )



from peft import prepare_model_for_kbit_training
from peft import LoraConfig, get_peft_model
import warnings
from transformers import (
    WhisperConfig,
    WhisperFeatureExtractor,
    WhisperForConditionalGeneration,
    WhisperProcessor,
    WhisperTokenizerFast,
    WhisperTokenizer,
    BitsAndBytesConfig
)
import torch
warnings.filterwarnings("ignore")

class WhisperModelPrep:
    """Facilitates the preparation of datasets for fine-tuning with the Whisper model,
    encompassing feature extraction, tokenization, and dataset processing.

    Attributes
    ----------
        dataset (DatasetDict): Dataset to be prepared for the Whisper model.
        model_id (str): Identifier for the Whisper model to be utilized.
        language_code (str): ISO code representing the dataset's language.
        processing_task (str): Specific task for the Whisper model to execute.

    """

    def __init__(
        self,
        model_id: str,
        language: str,
        processing_task: str,
        use_peft: bool,
    ):
        """Sets up the dataset and configuration for processing with the Whisper model.

        Parameters
        ----------
            dataset (DatasetDict): The dataset to be prepared.
            model_id (str, optional): Whisper model identifier.
            language_code (str, optional): Dataset language ISO code.
            processing_task (str, optional): Task for the Whisper model.

        """

        self.model_id = model_id
        self.language = language
        self.processing_task = processing_task
        self.use_peft = use_peft

    def initialize_feature_extractor(self) -> WhisperFeatureExtractor:
        """Creates and retrieves a feature extractor based on the Whisper model.

        Returns
        -------
            WhisperFeatureExtractor: Configured feature extractor for the model.

        """
        return WhisperFeatureExtractor.from_pretrained(
            self.model_id
        )

    def initialize_tokenizer(self) -> WhisperTokenizer:
        """Creates and retrieves a tokenizer for the Whisper model.

        Returns
        -------
            WhisperTokenizer: Configured tokenizer for the model.

        """
        return WhisperTokenizer.from_pretrained(
            self.model_id
        )

    def initialize_processor(self) -> WhisperProcessor:
        """Combines the feature extractor and tokenizer to create a processor for the Whisper model,
        facilitating seamless data preparation.

        Returns
        -------
            WhisperProcessor: Unified processor for the model.

        """
        return WhisperProcessor.from_pretrained(
            self.model_id, self.processing_task
        )

    def initialize_model(self) -> WhisperForConditionalGeneration:
        """Initializes and retrieves the Whisper model configured for conditional generation.

        This method sets up the Whisper model with specific configurations, ensuring it is
        ready for use in tasks such as transcription or translation, depending on the
        processing task specified during the class initialization.

        Returns
        -------
            WhisperForConditionalGeneration: The configured Whisper model ready for conditional generation tasks.

        """
        whisper_lang_code = convert_language_code(self.language)
        if self.use_peft:
            quantization_config = BitsAndBytesConfig(load_in_8bit=True)
            model = WhisperForConditionalGeneration.from_pretrained(
                self.model_id,
                quantization_config=quantization_config,
                device_map="auto",
            )
            model.config.forced_decoder_ids = None
            model.config.suppress_tokens = []
            model.config.use_cache = False
            model.generation_config.language = whisper_lang_code if self.processing_task == "transcribe" else "en"
            model.generation_config.task = self.processing_task
            model = prepare_model_for_kbit_training(model)
            config = LoraConfig(
                r=32,
                lora_alpha=64,
                target_modules=["q_proj", "v_proj"],
                lora_dropout=0.05,
                bias="none",
            )
            model = get_peft_model(model, config)
            model.print_trainable_parameters()
        else:
            print("PEFT optimization is not enabled.")
            model = WhisperForConditionalGeneration.from_pretrained(
                self.model_id,
                low_cpu_mem_usage = True
            )
            model.config.forced_decoder_ids = None
            model.config.suppress_tokens = []
            model.config.use_cache = False
            model.generation_config.language = whisper_lang_code if self.processing_task == "transcribe" else "en"
            model.generation_config.task = self.processing_task
            model = model.to("cuda") if torch.cuda.is_available() else model
        return model

# Define the mapping between FLEURS language codes and Whisper language tokens
fleurs_to_whisper = {
    "af_za": "af",  # Afrikaans
    "am_et": "am",  # Amharic
    "ar_eg": "ar",  # Arabic
    "as_in": "as",  # Assamese
    "az_az": "az",  # Azerbaijani
    "be_by": "be",  # Belarusian
    "bg_bg": "bg",  # Bulgarian
    "bn_in": "bn",  # Bengali
    "bs_ba": "bs",  # Bosnian
    "ca_es": "ca",  # Catalan
    "cs_cz": "cs",  # Czech
    "cy_gb": "cy",  # Welsh
    "da_dk": "da",  # Danish
    "de_de": "de",  # German
    "el_gr": "el",  # Greek
    "en_us": "en",  # English
    "es_es": "es",  # Spanish
    "et_ee": "et",  # Estonian
    "fa_ir": "fa",  # Persian
    "fi_fi": "fi",  # Finnish
    "fr_fr": "fr",  # French
    "ga_ie": "ga",  # Irish
    "gl_es": "gl",  # Galician
    "gu_in": "gu",  # Gujarati
    "he_il": "he",  # Hebrew
    "hi_in": "hi",  # Hindi
    "hr_hr": "hr",  # Croatian
    "hu_hu": "hu",  # Hungarian
    "hy_am": "hy",  # Armenian
    "id_id": "id",  # Indonesian
    "is_is": "is",  # Icelandic
    "it_it": "it",  # Italian
    "ja_jp": "ja",  # Japanese
    "jv_id": "jv",  # Javanese
    "ka_ge": "ka",  # Georgian
    "kk_kz": "kk",  # Kazakh
    "km_kh": "km",  # Khmer
    "kn_in": "kn",  # Kannada
    "ko_kr": "ko",  # Korean
    "lo_la": "lo",  # Lao
    "lt_lt": "lt",  # Lithuanian
    "lv_lv": "lv",  # Latvian
    "mk_mk": "mk",  # Macedonian
    "ml_in": "ml",  # Malayalam
    "mn_mn": "mn",  # Mongolian
    "mr_in": "mr",  # Marathi
    "ms_my": "ms",  # Malay
    "my_mm": "my",  # Burmese
    "ne_np": "ne",  # Nepali
    "nl_nl": "nl",  # Dutch
    "no_no": "no",  # Norwegian
    "or_in": "or",  # Odia
    "pa_in": "pa",  # Punjabi
    "pl_pl": "pl",  # Polish
    "pt_br": "pt",  # Portuguese
    "ro_ro": "ro",  # Romanian
    "ru_ru": "ru",  # Russian
    "si_lk": "si",  # Sinhala
    "sk_sk": "sk",  # Slovak
    "sl_si": "sl",  # Slovenian
    "sq_al": "sq",  # Albanian
    "sr_rs": "sr",  # Serbian
    "sv_se": "sv",  # Swedish
    "sw_ke": "sw",  # Swahili
    "ta_in": "ta",  # Tamil
    "te_in": "te",  # Telugu
    "th_th": "th",  # Thai
    "tl_ph": "tl",  # Filipino
    "tr_tr": "tr",  # Turkish
    "uk_ua": "uk",  # Ukrainian
    "ur_pk": "ur",  # Urdu
    "vi_vn": "vi",  # Vietnamese
    "zh_cn": "zh",  # Chinese
}
def convert_language_code(language_code):
    if language_code in fleurs_to_whisper.values():
        return language_code
    return fleurs_to_whisper.get(language_code, None)

#################################


def load_model_ps(
        model_name_or_path,
        token,
        model_revision="main",
        cache_dir=None,
        feature_extractor_name=None,
        config_name=None,
        tokenizer_name=None,
        processor_name=None,
        use_fast_tokenizer=True,
        subfolder="",
        attn_implementation=None,
        language=None,
        dtype="bfloat16",
        return_timestamps=False,
        task="transcribe",
):
    if dtype == "float16":
        # mixed_precision = "fp16"
        torch_dtype = torch.float16
    elif dtype == "bfloat16":
        # mixed_precision = "bf16"
        torch_dtype = torch.bfloat16
    else:
        # mixed_precision = "no"
        torch_dtype = torch.float32

    config = WhisperConfig.from_pretrained(
            (config_name if config_name else model_name_or_path),
            cache_dir=cache_dir,
            revision=model_revision,
            token=token,
        )
    feature_extractor = WhisperFeatureExtractor.from_pretrained(
        (feature_extractor_name if feature_extractor_name else model_name_or_path),
        cache_dir=cache_dir,
        revision=model_revision,
        token=token,
    )
    tokenizer = WhisperTokenizerFast.from_pretrained(
        (tokenizer_name if tokenizer_name else model_name_or_path),
        cache_dir=cache_dir,
        use_fast=use_fast_tokenizer,
        revision=model_revision,
        token=token,
    )
    processor = WhisperProcessor.from_pretrained(
        (processor_name if processor_name else model_name_or_path),
        cache_dir=cache_dir,
        revision=model_revision,
        token=token,
    )

    model = WhisperForConditionalGeneration.from_pretrained(
        model_name_or_path,
        config=config,
        cache_dir=cache_dir,
        revision=model_revision,
        subfolder=subfolder,
        token=token,
        low_cpu_mem_usage=True,
        torch_dtype=torch_dtype,
        attn_implementation=attn_implementation,
    )
    model.eval()

    if model.config.decoder_start_token_id is None:
        raise ValueError("Make sure that `config.decoder_start_token_id` is correctly defined")

    return_timestamps = return_timestamps
    if hasattr(model.generation_config, "is_multilingual") and model.generation_config.is_multilingual:
        is_multilingual = True
        # We need to set the language and task ids for multilingual checkpoints
        tokenizer.set_prefix_tokens(
            language=language, task=task, predict_timestamps=return_timestamps
        )
    elif language is not None:
        raise ValueError(
            "Setting language token for an English-only checkpoint is not permitted. The language argument should "
            "only be set for multilingual checkpoints."
        )
    else:
        is_multilingual = False
    return model, feature_extractor, tokenizer, processor, is_multilingual



from transformers import (
    WhisperFeatureExtractor,
    WhisperTokenizer,
    WhisperProcessor,
    WhisperForConditionalGeneration,
)
from transformers.models.whisper.english_normalizer import BasicTextNormalizer, EnglishTextNormalizer
from huggingface_hub import create_repo, get_full_repo_name, Repository
from pathlib import Path
import numpy as np
import datasets
import os
from datasets import (
    DatasetDict,
    IterableDatasetDict,
)
from soundfile import LibsndfileError
from datasets.arrow_dataset import table_iter
from accelerate.logging import get_logger
from .load_data import Dataset
from .whisper_model_prep import WhisperModelPrep
from .audio_data_processor import AudioDataProcessor
from typing import Tuple
import warnings
warnings.filterwarnings("ignore")

logger = get_logger(__name__)



class DataPrep:
    """

    A class to encapsulate preparing the speech dataset for model training, including dataset loading, cleaning, and preprocessing tasks like feature extraction and tokenization. .

    """

    def __init__(
        self,
        huggingface_token: str,
        dataset_name: str,
        language_abbr: str,
        model_id: str,
        processing_task: str,
        use_peft: bool,
    ):
        """
        Initializes the Trainer with the necessary configuration and loads the evaluation metric.

        Parameters:
            huggingface_token (str): Hugging Face API token for authenticated access.
            dataset_name (str): Name of the dataset to be downloaded from Hugging Face.
            language_abbr (str): Language abbreviation for the dataset.
            model_id (str): Model ID for the model to be used in training.
            processing_task (str): The processing task to be performed (e.g., "transcribe").
        """
        self.huggingface_token = huggingface_token
        self.dataset_name = dataset_name
        self.language_abbr = language_abbr
        self.model_id = model_id
        self.processing_task = processing_task
        self.use_peft = use_peft
        self.model_prep = WhisperModelPrep(
            language = self.language_abbr,
            model_id=self.model_id,
            processing_task=self.processing_task,
            use_peft=self.use_peft,
        )
        self.data_loader = Dataset(
            self.huggingface_token, self.dataset_name, self.language_abbr
        )

    def prepare_model(
        self,
    ) -> Tuple[
        WhisperTokenizer,
        WhisperFeatureExtractor,
        WhisperProcessor,
        WhisperForConditionalGeneration,
    ]:
        """Initializes and sets up the necessary components for model training, including the tokenizer,
        feature extractor, feature processor, and the model itself.

        Returns:
        tuple: A tuple containing four elements in the order of tokenizer, feature extractor,
               feature processor, and the model. Each element is an initialized instance of its
               respective class, ready for use in the training pipeline.
        """

        self.tokenizer = self.model_prep.initialize_tokenizer()
        self.feature_extractor = self.model_prep.initialize_feature_extractor()
        self.feature_processor = self.model_prep.initialize_processor()
        self.model = self.model_prep.initialize_model()
        return (
            self.tokenizer,
            self.feature_extractor,
            self.feature_processor,
            self.model,
        )

    def load_dataset(
        self,  
        feature_extractor: WhisperFeatureExtractor, 
        tokenizer: WhisperTokenizer, 
        processor: WhisperProcessor,
        streaming: bool = True,
        train_num_samples: int = None,
        test_num_samples: int = None) -> dict:
        """
        Retrieves and preprocesses the specified dataset for model training and evaluation.

        Parameters:
        feature_extractor (PreTrainedFeatureExtractor):
            The feature extractor instance to be used for audio data preprocessing.
        tokenizer (PreTrainedTokenizer):
            The tokenizer instance for processing textual data associated with the audio samples.
        processor (WhisperProcessor):
            The processor instance for processing the audio samples.
        streaming (bool):
            If True, the datasets will be streamed, allowing for loading large datasets without requiring them to fit into memory.
            If False, the entire dataset will be downloaded before processing.
        train_num_samples (int, optional):
            The maximum number of training samples to load from each dataset.
            For example, if train_num_samples = 100, then 100 samples will be loaded from each dataset's training split.
            If None, the entire training split will be loaded.
        test_num_samples (int, optional):
            The maximum number of test samples to load from each dataset.
            For example, if test_num_samples = 100, then 100 samples will be loaded from each dataset's test split.
            If None, the entire test split will be loaded.

        Returns:
            dict: A dictionary containing the preprocessed 'train' and 'test' splits of the dataset,
                        ready for use in model training and evaluation. Each split has been cleaned and
                        processed to include only the necessary features for model input.
        """

        dataset = self.data_loader.load_dataset(
            streaming,
            train_num_samples = train_num_samples,
            test_num_samples = test_num_samples
        )
        print(dataset)
        # train_count, test_count = self.data_loader.count_examples(dataset)
        # print(f"Training dataset size: {train_count}")
        # print(f"Test dataset size: {test_count}")
        processor = AudioDataProcessor(dataset, feature_extractor, tokenizer, processor)
        dataset['train']= dataset['train'].map(processor.resampled_dataset, remove_columns=list(next(iter(dataset['train'])).keys()))
        dataset['test']= dataset['test'].map(processor.resampled_dataset)
        return dataset


##########################

def preprocess_datasets(
    model,
    feature_extractor,
    tokenizer,
    is_multilingual,
    language,
    data_splits,
    raw_datasets,
    accelerator,
    token,
    dataset_name=None,
    max_duration_in_seconds=30.0,
    max_label_length=256,
    audio_column_name="audio",
    preprocessing_batch_size=500,
    preprocessing_num_workers=None,
    dataloader_num_workers=0,
    text_column_name="sentence",
    id_column_name="id",
    speaker_id_column_name=None,
    max_samples_per_split=None,
    streaming=False,
    concatenate_audio=True,
    preprocessing_only=False,
    output_dir=None,
    push_to_hub=True,
    hub_model_id=None
):
    max_input_length = int(max_duration_in_seconds * feature_extractor.sampling_rate)
    max_label_length = (
        max_label_length if max_label_length is not None else model.config.max_length
    )
    audio_column_name = audio_column_name
    sampling_rate = feature_extractor.sampling_rate

    preprocessing_batch_size = preprocessing_batch_size
    num_workers = preprocessing_num_workers
    dataloader_num_workers = dataloader_num_workers

    text_column_name = text_column_name
    model_input_name = feature_extractor.model_input_names[0]
    id_column_name = id_column_name
    speaker_id_column_name = speaker_id_column_name
    normalizer = (
        BasicTextNormalizer()
        if language is not None
        else EnglishTextNormalizer(tokenizer.english_spelling_normalizer)
    )

    timestamp_position = 3 if is_multilingual else 1
    decoder_prev_token_id = tokenizer.convert_tokens_to_ids("<|startofprev|>")
    decoder_eot_token_id = tokenizer.eos_token_id

    if max_samples_per_split is not None:
        for split in data_splits:
            raw_datasets[split] = (
                raw_datasets[split].take(max_samples_per_split)
                if streaming
                else raw_datasets[split].select(range(max_samples_per_split))
            )

    if speaker_id_column_name is not None:
        raw_datasets = raw_datasets.sort(speaker_id_column_name)


    def concatenate_dataset(batch):
        audio_arrays, texts, speaker_ids = [], [], []

        # skip corrupted samples
        for row in table_iter(batch.pa_table, batch_size=1):
            row = batch.formatter.format_row(row)
            try:
                sample_audio = row[audio_column_name]['array']
                sample_text = row[text_column_name]
                sample_speaker_id = row[speaker_id_column_name] if speaker_id_column_name else None
            except LibsndfileError:
                logger.warning(f"{row[id_column_name]} is corrupted! Skipping sample.")
                continue
            audio_arrays.append(sample_audio)
            texts.append(sample_text)
            speaker_ids.append(sample_speaker_id)

        # initialize concatenations
        concat_audio = [audio_arrays[0]]
        concat_text = [texts[0]]
        concat_speaker_id = [speaker_ids[0]]
        condition_on_prev = [0]

        for audio_array, text, speaker_id in zip(audio_arrays[1:], texts[1:], speaker_ids[1:]):
            is_same_speaker = speaker_id == concat_speaker_id[-1]
            is_concatenable = len(audio_array) + len(concat_audio[-1]) <= max_input_length
            if is_same_speaker and is_concatenable:
                # inplace concatenation
                concat_audio[-1] = np.append(concat_audio[-1], audio_array)
                concat_text[-1] = concat_text[-1] + " " + text
            else:
                concat_audio.append(audio_array)
                concat_text.append(text)
                concat_speaker_id.append(speaker_id)
                condition_on_prev.append(1 if is_same_speaker else 0)

        batch[audio_column_name] = [{"array": array, "sampling_rate": sampling_rate} for array in concat_audio]
        batch[text_column_name] = concat_text
        batch[id_column_name] = concat_speaker_id
        batch["condition_on_prev"] = condition_on_prev

        return batch

    raw_datasets_features = list(next(iter(raw_datasets.values())).features.keys())
    if concatenate_audio and not streaming:
        with accelerator.main_process_first():
            raw_datasets = raw_datasets.map(
                concatenate_dataset,
                batched=True,
                batch_size=preprocessing_batch_size,
                num_proc=num_workers,
                remove_columns=set(raw_datasets_features)
                               - {audio_column_name, text_column_name, id_column_name, "condition_on_prev"},
                desc="Concatenating dataset...",
            )

        raw_datasets = raw_datasets.cast_column(
            audio_column_name, datasets.features.Audio(sampling_rate=sampling_rate)
        )
        pretty_name = dataset_name.split("/")[-1]

        def postprocess_ids(speaker_ids, indices):
            speaker_ids_formatted = []
            for speaker, idx in zip(speaker_ids, indices):
                formatted_idx = f"{pretty_name}-{speaker}-{idx}" if speaker is not None else f"{pretty_name}-{idx}"
                speaker_ids_formatted.append(formatted_idx)
            return {id_column_name: speaker_ids_formatted}

        with accelerator.main_process_first():
            raw_datasets = raw_datasets.map(
                postprocess_ids,
                input_columns=[id_column_name],
                with_indices=True,
                desc="Setting sample idxs...",
                batched=True,
                batch_size=preprocessing_batch_size,
                num_proc=num_workers,
            )
    elif concatenate_audio and streaming:
        raise ValueError(
            "Streaming mode is not yet compatible with concatenating audios to `max_duration_in_seconds`."
            "Either set `--streaming=False` and download the audios locally, or open an issue on the Distil-Whisper repo to request this feature."
        )

    def prepare_dataset(batch):
        # process audio
        sample = batch[audio_column_name]
        inputs = feature_extractor(sample["array"], sampling_rate=sample["sampling_rate"])
        # process audio length
        batch[model_input_name] = inputs.get(model_input_name)[0]

        # process targets
        input_str = batch[text_column_name]
        batch["labels"] = tokenizer(input_str, max_length=max_label_length, truncation=True).input_ids
        return batch

    raw_datasets_features = list(next(iter(raw_datasets.values())).features.keys())
    file_ids_dataset = IterableDatasetDict() if streaming else DatasetDict()
    for split in raw_datasets:
        file_ids_dataset[split] = raw_datasets[split][id_column_name]
    if streaming:
        with accelerator.main_process_first():
            vectorized_datasets = raw_datasets.map(prepare_dataset, remove_columns=raw_datasets_features)
    else:
        with accelerator.main_process_first():
            vectorized_datasets = raw_datasets.map(
                prepare_dataset,
                remove_columns=raw_datasets_features,
                num_proc=num_workers,
                desc="preprocess dataset",
            )

    # for large datasets it is advised to run the preprocessing on a
    # single machine first with `args.preprocessing_only` since there will mostly likely
    # be a timeout when running the script in distributed mode.
    # In a second step `args.preprocessing_only` can then be set to `False` to load the
    # cached dataset
    if preprocessing_only:
        cache = {k: v.cache_files for k, v in vectorized_datasets.items()}
        logger.info(f"Data preprocessing finished. Files cached at {cache}.")
        return

    if streaming and dataloader_num_workers > 0:
        logger.warning(
            "Using multiple dataloader num workers with streaming mode will result in different shards of "
            "data being transcribed in parallel. This is not advised if you want to preserve the order of the "
            "audio-text data."
        )

    # Handle the repository creation
    repo_name = None
    output_dir = output_dir
    if accelerator.is_main_process:
        if push_to_hub:
            if hub_model_id is None:
                repo_name = get_full_repo_name(
                    Path(output_dir).absolute().name,
                    token=token,
                )
            else:
                repo_name = hub_model_id
            create_repo(repo_name, repo_type="dataset", exist_ok=True, token=token)
            Repository(
                output_dir,
                clone_from=repo_name,
                token=token,
                repo_type="dataset",
            )

            # Ensure large txt files can be pushed to the Hub with git-lfs
            with open(os.path.join(output_dir, ".gitattributes"), "r+") as f:
                git_lfs_extensions = f.read()
                if "*.csv" not in git_lfs_extensions:
                    f.write("*.csv filter=lfs diff=lfs merge=lfs -text")

        elif output_dir is not None:
            # this is where we'll save our transcriptions
            os.makedirs(output_dir, exist_ok=True)

    accelerator.wait_for_everyone()
    return vectorized_datasets, file_ids_dataset, decoder_eot_token_id, decoder_prev_token_id, timestamp_position, repo_name, normalizer




import torch
from dataclasses import dataclass
from typing import Any, Dict, List, Union, Optional
import numpy as np
import warnings

warnings.filterwarnings("ignore")

@dataclass
class DataCollatorSpeechSeq2SeqWithPadding:
    """
    A data collator that dynamically pads a batch of inputs for speech-to-text sequence-to-sequence models.

    Attributes:
        processor (Any): The processor combining feature extraction and tokenization,
                         used for preparing model inputs and labels.
    """

    processor: Any

    def __call__(
        self, features: List[Dict[str, Union[List[int], torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        """
        Processes and pads a batch of input features and labels for training or evaluation.

        Args:
            features (List[Dict[str, Union[List[int], torch.Tensor]]]): A list of feature dictionaries,
                each containing 'input_features' for model inputs and 'labels' for expected outputs.

        Returns:
            Dict[str, torch.Tensor]: A dictionary of padded tensors including 'input_features' for the model inputs
                                     and 'labels' for the model outputs, with padding applied according to the processor's
                                     configurations.
        """
        input_features = [
            {"input_features": feature["input_features"]} for feature in features
        ]
        batch = self.processor.feature_extractor.pad(
            input_features, return_tensors="pt"
        )

        label_features = [{"input_ids": feature["labels"]} for feature in features]
        labels_batch = self.processor.tokenizer.pad(label_features, return_tensors="pt")

        labels = labels_batch["input_ids"].masked_fill(
            labels_batch.attention_mask.ne(1), -100
        )

        if (labels[:, 0] == self.processor.tokenizer.bos_token_id).all().cpu().item():
            labels = labels[:, 1:]

        batch["labels"] = labels

        return batch


@dataclass
class DataCollatorSpeechSeq2SeqWithPaddingPS:
    """
    Data collator that will dynamically pad the inputs received.
    Args:
        processor ([`Wav2Vec2Processor`])
            The processor used for proccessing the data.
        decoder_start_token_id (:obj: `int`)
            The start-of-sequence token id of the decoder.
        input_padding (:obj:`bool`, :obj:`str` or :class:`~transformers.tokenization_utils_base.PaddingStrategy`, `optional`, defaults to :obj:`True`):
            Select a strategy to pad the returned input sequences (according to the model's padding side and padding index)
            among:
            * :obj:`True` or :obj:`'longest'`: Pad to the longest sequence in the batch (or no padding if only a single
              sequence if provided).
            * :obj:`'max_length'`: Pad to a maximum length specified with the argument :obj:`max_length` or to the
              maximum acceptable input length for the model if that argument is not provided.
            * :obj:`False` or :obj:`'do_not_pad'` (default): No padding (i.e., can output a batch with sequences of
              different lengths).
        target_padding (:obj:`bool`, :obj:`str` or :class:`~transformers.tokenization_utils_base.PaddingStrategy`, `optional`, defaults to :obj:`True`):
            Select a strategy to pad the returned target sequences (according to the model's padding side and padding index).
            See above for details.
        max_target_length (:obj:`int`, `optional`):
            Maximum length of the ``labels`` of the returned list and optionally padding length (see above).
    """

    processor: Any
    decoder_start_token_id: int
    input_padding: Union[bool, str] = "max_length"
    target_padding: Union[bool, str] = "max_length"
    max_target_length: Optional[int] = None

    def __call__(self, features: List[Dict[str, Union[List[int], np.ndarray]]]) -> Dict[str, np.ndarray]:
        # split inputs and labels since they have to be of different lengths and need
        # different padding methods
        model_input_name = self.processor.model_input_names[0]

        # dataloader returns a list of features which we convert to a dict
        input_features = {model_input_name: [feature[model_input_name] for feature in features]}
        label_features = {"input_ids": [feature["labels"] for feature in features]}

        # reformat list to dict and set to pytorch format
        batch = self.processor.feature_extractor.pad(
            input_features,
            padding=self.input_padding,
            return_tensors="pt",
        )

        labels_batch = self.processor.tokenizer.pad(
            label_features,
            max_length=self.max_target_length,
            padding=self.target_padding,
            return_tensors="pt",
        )

        # replace padding with -100 to ignore correctly when computing the loss
        labels = labels_batch["input_ids"].masked_fill(labels_batch.attention_mask.ne(1), -100)

        # if bos token is appended in previous tokenization step,
        # cut bos token here as it's append later anyways
        if (labels[:, 0] == self.decoder_start_token_id).all().cpu().item():
            labels = labels[:, 1:]

        batch["labels"] = labels
        return batch


from datasets import load_dataset, IterableDatasetDict
import warnings
from datasets import DatasetDict
from huggingface_hub import HfFolder
warnings.filterwarnings("ignore")

class Dataset:
    """
    Manages loading of the datasets from Hugging Face's Common Voice repository.

    Attributes:
        huggingface_token (str): Hugging Face API token for read authenticated access.
        dataset_name (str): Name of the dataset to be downloaded from Hugging Face.
        language_abbr (str): Abbreviation of the language for the dataset.
    """

    def __init__(self, huggingface_token: str, dataset_name: str, language_abbr: str):
        """
        Initializes the DatasetManager with necessary details for dataset operations.

        Parameters:
            huggingface_token (str): Hugging Face API token.
            dataset_name (str): Name of the dataset.
            language_abbr (str): Language abbreviation for the dataset.
        """
        self.huggingface_token = huggingface_token
        self.dataset_name = dataset_name
        self.language_abbr = language_abbr

    def load_dataset(self, streaming: bool = True, train_num_samples: int = None, test_num_samples: int = None) -> dict:
        """Load datasets for each language abbreviation and concatenate train/test splits.

        Parameters:
        streaming (bool):
            If True, the datasets will be streamed, allowing for loading large datasets without requiring them to fit into memory.
            If False, the entire dataset will be downloaded before processing.
        train_num_samples (int, optional): The maximum number of training samples to load from each dataset.
            For example, if train_num_samples = 100, then 100 samples will be loaded from each dataset's training split.
            If None, the entire training split will be loaded.
        test_num_samples (int, optional): The maximum number of test samples to load from each dataset.
            For example, if test_num_samples = 100, then 100 samples will be loaded from each dataset's test split.
            If None, the entire test split will be loaded.

        Returns:
            dict: A dictionary containing concatenated train and test splits for each language.
        """
        data = {}

        train_dataset = load_dataset(
            self.dataset_name,
            self.language_abbr,
            split="train",
            streaming=streaming,
            token=self.huggingface_token,
            trust_remote_code=True,
        )
        test_dataset = load_dataset(
            self.dataset_name,
            self.language_abbr,
            split="test",
            streaming=streaming,
            token=self.huggingface_token,
            trust_remote_code=True,
        )

        if streaming:
            train_split = train_dataset.take(train_num_samples) if train_num_samples else train_dataset
        else:
            train_split = (
                train_dataset
                if not train_num_samples or len(train_dataset) < train_num_samples
                else train_dataset.select(range(train_num_samples))
            )

        if streaming:
            test_split = test_dataset.take(test_num_samples) if test_num_samples else test_dataset
        else:
            test_split = (
                test_dataset
                if not test_num_samples or len(test_dataset) < test_num_samples
                else test_dataset.select(range(test_num_samples))
            )

        data["train"] = train_split
        data["test"] = test_split

        return data

    @staticmethod
    def count_examples(dataset: dict) -> tuple:
        """
        Count the number of examples in the dataset.

        Args:
            dataset (IterableDataset): The dataset to count examples from.

        Returns:
            train_samples: The number of training examples in the dataset.
            test_samples: The number of test examples in the dataset.
        """
        train_samples = list(dataset["train"])
        test_samples = list(dataset["test"])
        train_samples = len(train_samples)
        test_samples = len(test_samples)
        return train_samples, test_samples


def load_and_validate_ps_datasets(
        token,
        dataset_split_name,
        accelerator,
        dataset_name,
        dataset_config_name,
        dataset_cache_dir,
        streaming,
        preprocessing_num_workers=None,
        audio_column_name="audio",
        text_column_name="sentence",

    ):
    raw_datasets = IterableDatasetDict() if streaming else DatasetDict()
    token = token if token is not None else HfFolder().get_token()

    data_splits = dataset_split_name.split("+")

    for split in data_splits:
        with accelerator.main_process_first():
            raw_datasets[split] = load_dataset(
                dataset_name,
                dataset_config_name,
                split=split,
                cache_dir=dataset_cache_dir,
                token=token,
                streaming=streaming,
                num_proc=preprocessing_num_workers if not streaming else None,
                trust_remote_code=True
            )

    if audio_column_name not in next(iter(raw_datasets.values())).column_names:
        raise ValueError(
            f"--audio_column_name '{audio_column_name}' not found in dataset"
            f" '{dataset_name}'. Make sure to set `--audio_column_name` to"
            " the correct audio column - one of"
            f" {', '.join(next(iter(raw_datasets.values())).column_names)}."
        )

    if text_column_name not in next(iter(raw_datasets.values())).column_names:
        raise ValueError(
            f"--text_column_name {text_column_name} not found in dataset"
            f" '{dataset_name}'. Make sure to set `--text_column_name` to the"
            " correct text column - one of"
            f" {', '.join(next(iter(raw_datasets.values())).column_names)}."
        )

    return raw_datasets, data_splits



import os
import unittest
from src.training.audio_data_processor import AudioDataProcessor
from src.training.whisper_model_prep import WhisperModelPrep
from src.training.load_data import Dataset
from dotenv import load_dotenv
load_dotenv()

class TestAudioDataProcessor(unittest.TestCase):
    """
    Unit tests for the AudioDataProcessor class.
    """

    def setUp(self):
        """
        Set up test environment.
        """
        # Load dataset
        self.data_loader = Dataset(
            huggingface_token = os.environ.get("HF_TOKEN"),
            dataset_name="mozilla-foundation/common_voice_16_1",
            language_abbr="af"
        )
        self.dataset_streaming = self.data_loader.load_dataset(streaming=True, train_num_samples=10, test_num_samples=10)
        self.dataset_batch = self.data_loader.load_dataset(streaming=False, train_num_samples=10, test_num_samples=10)

        has_train_sample = any(True for _ in self.dataset_streaming["train"])
        assert has_train_sample, "Train dataset is empty!"

        has_test_sample = any(True for _ in self.dataset_streaming["test"])
        assert has_test_sample, "Test dataset is empty!"

        has_train_sample = any(True for _ in self.dataset_batch["train"])
        assert has_train_sample, "Train dataset is empty!"

        has_test_sample = any(True for _ in self.dataset_batch["test"])
        assert has_test_sample, "Test dataset is empty!"

        # Initialize model preparation
        self.model_prep = WhisperModelPrep(
            language = "af",
            model_id="openai/whisper-tiny",
            processing_task="transcribe",
            use_peft=False
        )

        # Initialize tokenizer, feature extractor, and feature processor
        self.tokenizer = self.model_prep.initialize_tokenizer()
        self.feature_extractor = self.model_prep.initialize_feature_extractor()
        self.feature_processor = self.model_prep.initialize_processor()

        # Initialize AudioDataProcessor
        self.processor = AudioDataProcessor(
            dataset=self.dataset_streaming,
            feature_extractor=self.feature_extractor,
            tokenizer=self.tokenizer,
            feature_processor=self.feature_processor
        )

        self.processor_batch = AudioDataProcessor(
            dataset=self.dataset_batch,
            feature_extractor=self.feature_extractor,
            tokenizer=self.tokenizer,
            feature_processor=self.feature_processor
        )

    def test_resampled_dataset(self):
        """
        Test the resampled_dataset method.
        """
        # Arrange
        sample_dataset = self.dataset_streaming

        # Act & Assert
        for split, samples in sample_dataset.items():
            for sample in samples:
                resampled_data = self.processor.resampled_dataset(sample)
                self.assertIn("input_features", resampled_data)
                self.assertIn("labels", resampled_data)
                self.assertEqual(resampled_data["audio"]["sampling_rate"], 16000)

    def test_resampled_dataset_batch(self):
        """
        Test the resampled_dataset method.
        """
        # Arrange
        sample_dataset = self.dataset_batch

        # Act & Assert
        for split, samples in sample_dataset.items():
            for sample in samples:
                resampled_data = self.processor_batch.resampled_dataset(sample)
                self.assertIn("input_features", resampled_data)
                self.assertIn("labels", resampled_data)
                self.assertEqual(resampled_data["audio"]["sampling_rate"], 16000)

if __name__ == '__main__':
    unittest.main()



import argparse
from .data_prep import DataPrep
from .model_trainer import Trainer
import warnings

warnings.filterwarnings("ignore")

def parse_args():
    parser = argparse.ArgumentParser(
        description="Run the training orchestrator with specified parameters."
    )
    parser.add_argument(
        "--huggingface_token",
        type=str,
        required=True,
        help="Hugging Face API token for write authenticated access.",
    )
    parser.add_argument(
        "--dataset_name",
        type=str,
        default="mozilla-foundation/common_voice_16_1",
        help="Name of the dataset to be downloaded from Hugging Face.",
    )
    parser.add_argument(
        "--train_num_samples",
        type=int,
        default=None,
        help="Number of training samples in the dataset",
    )
    parser.add_argument(
        "--test_num_samples",
        type=int,
        default=None,
        help="Number of testing samples in the dataset",
    )
    parser.add_argument(
        "--streaming",
        type=bool,
        default=True,
        help="Load dataset in streaming or Batch mode",
    )
    parser.add_argument(
        "--language_abbr",
        required=True,
        help="Abbreviation of the language(s) for the dataset.",
    )
    parser.add_argument(
        "--model_id",
        type=str,
        default="openai/whisper-small",
        help="Model ID for the model to be used in training.",
    )
    parser.add_argument(
        "--processing_task",
        type=str,
        default="transcribe",
        help="The processing task to be performed.",
    )
    parser.add_argument(
        "--wandb_api_key",
        type=str,
        help="The wandb.ai api key for monitoring training runs, signup and generate an api key.",
    )
    parser.add_argument(
        "--use_peft",
        type=bool,
        help="True to train your model using PEFT method, False for full finetuning",
    )
    parser.add_argument(
        "--max_steps",
        type=int,
        help="The maximum number of training steps. Defaults to 100",
    )
    parser.add_argument(
        "--train_batch_size",
        type=int,
        help="The batch size per GPU during training. Defaults to 8",
    )
    parser.add_argument(
        "--eval_batch_size",
        type=int,
        help="The batch size per GPU during evaluation. Defaults to 8",
    )
    parser.add_argument(
        "--save_eval_logging_steps",
        type=int, 
        help="The number of training steps before saving the model, evaluating the model and logging the training info. Defaults to 25",
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    process = DataPrep(
        huggingface_token=args.huggingface_token,
        dataset_name=args.dataset_name,
        language_abbr=args.language_abbr,
        model_id=args.model_id,
        processing_task=args.processing_task,
        use_peft=args.use_peft,
    )
    tokenizer, feature_extractor, feature_processor, model = process.prepare_model()

    dataset = process.load_dataset(feature_extractor, 
                                   tokenizer, 
                                   feature_processor,
                                   streaming = args.streaming,
                                   train_num_samples = args.train_num_samples,
                                   test_num_samples = args.test_num_samples)
    trainer = Trainer(

        huggingface_token=args.huggingface_token,
        model_id=args.model_id,
        dataset=dataset,
        language = args.language_abbr,
        model=model,
        feature_processor=feature_processor,
        feature_extractor=feature_extractor,
        tokenizer=tokenizer,
        wandb_api_key=args.wandb_api_key,
        use_peft=args.use_peft,
        processing_task=args.processing_task
    )
    trainer.train(
        max_steps=args.max_steps,
        per_device_train_batch_size=args.train_batch_size,
        per_device_eval_batch_size=args.eval_batch_size,
        save_steps=args.save_eval_logging_steps,
        eval_steps=args.save_eval_logging_steps,
        logging_steps=args.save_eval_logging_steps
    )
